/* ====================================================================
 *  Copyright 2003-2004 Paul Querna and Axel Grossklaus
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

/*
 * mod_authn_dbi
 * 
 * Paul Querna
 * Axel Grossklaus 
 *
 */

#include "../include/mod_authn_dbi_config.h"

#define MOD_AUTHN_DBI_VERSION VERSION

#include "apr_lib.h"
#define APR_WANT_STRFUNC
#include "apr_want.h"
#include "apr_strings.h"
#include "apr_md5.h"            /* for apr_password_validate */
#include "apr_sha1.h"
#include "apr_reslist.h"
#include "apr_hash.h"

#include "ap_provider.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"       /* for ap_hook_(check_user_id | auth_checker) */
#include "mod_auth.h"

#include  "util_md5.h"          /* for ap_md5 */

#include <dbi/dbi.h>

#define DFLT_DBI_NAME "AuthDB"
#define DFLT_DBI_HOST "localhost"
#define DFLT_DBI_DRIVER "mysql"
#define DFLT_DBI_USER "root"
#define DFLT_DBI_PASS "" /* setting this to NULL triggers a bug in libdbi which causes a segfault. mysql docs
                          * say this must be set to NULL in order to login without a password. fortunately, 
                          * setting "" works as well. the bug is reported to the libdbi maintainers. we will
                          * change this back when the bug is fixed.
                          *
                          */
#define DFLT_DBI_TABLE "Users"
#define DFLT_USERNAME_FIELD "Username"
#define DFLT_PASSWORD_FIELD "Password"
#define DFLT_PASSWORD_QUERY NULL
#define DFLT_PASSWORD_FORMAT plain
#define DFLT_ACTIVE_FIELD NULL
#define DFLT_CONN_MIN (1)
#define DFLT_CONN_SOFT (5)
#define DFLT_CONN_MAX (25)
#define DFLT_CONN_TTL (600)
#define DFLT_OPTIONS (0)

#ifndef DBI_HARD_MAX_CONNS
#define DBI_HARD_MAX_CONNS (255)
#endif

/* do NOT set this to the empty string ""! */
#define DBI_EMPTY_PASSWORD "::"

#define DBI_OPTION_EMPTY_PASSWORDS 0x01

typedef enum pw_format_enum
{
    plain,
    apr,                        /* handles crypt, sha1 & md5 through apr_password_validate() */
    aprdig
/*  
 *  Possible other formats to consider:
 *  pw_mysql
 *  pw_blowfish
 *  pw_md5
 *  ntlm
 */
} pw_format;

static int dbi_conn_count = 0;

static apr_hash_t *authn_dbi_config;

enum
{
    CONF_DBI_DRIVER,
    CONF_DBI_DRIVER_DIR,
    CONF_DBI_HOST,
    CONF_DBI_USERNAME,
    CONF_DBI_PASSWORD,
    CONF_DBI_NAME,
    CONF_DBI_TABLE,
    CONF_DBI_USERNAME_FIELD,
    CONF_DBI_PASSWORD_FIELD,
    CONF_DBI_PASSWORD_QUERY,
    CONF_DBI_PASSWORD_FORMAT,
    CONF_DBI_IS_ACTIVE_FIELD,
    CONF_DBI_CONN_MIN,
    CONF_DBI_CONN_SOFTMAX,
    CONF_DBI_CONN_HARDMAX,
    CONF_DBI_CONN_TTL,
    CONF_DBI_OPTIONS
};

typedef struct dbi_dconfig
{
    const char *id;
} dbi_dconfig;

typedef struct dbi_gconfig
{
    const char *driverdir;
} dbi_gconfig;

/* This might be a little clumsy but should do for now. */
static dbi_gconfig dbi_global_config = { NULL };

typedef struct dbi_config_rec_struct
{
    const char *dbi_name;
    const char *dbi_user;
    const char *dbi_pass;
    const char *dbi_driver;
    const char *dbi_host;
    const char *dbi_table;
    const char *username_field;
    const char *password_field;
    const char *password_query;
    pw_format password_format;
    const char *isactive_field;
    int conn_min;
    int conn_soft;
    int conn_max;
    int conn_ttl;
    apr_uint32_t options;
} dbi_config_rec;

typedef struct dbi_config_struct
{
    const char name;
    apr_reslist_t *pool;
    dbi_config_rec rec;
} dbi_config;

typedef struct dbi_rest_struct
{
    dbi_conn *conn;
} dbi_rest;

typedef const char *conn_id;

static apr_status_t safe_dbi_new_conn(void **resource, void *params,
                                      apr_pool_t * r)
{
    apr_status_t rv = APR_SUCCESS;
    dbi_config_rec *conf = params;
    int err_num = 0;
    const char *err_str;
    const char *host = conf->dbi_host;
    const char *driver = conf->dbi_driver;
    const char *name = conf->dbi_name;
    const char *user = conf->dbi_user;
    const char *pwd = conf->dbi_pass;
    dbi_rest *myres;

    dbi_conn_count++;

    if (DBI_HARD_MAX_CONNS > dbi_conn_count) {

        myres = apr_palloc(r, sizeof(*myres));

        myres->conn = dbi_conn_new(driver);
        if (myres->conn == NULL) {
            ap_log_perror(APLOG_MARK, APLOG_EMERG, 0, r,
                          "[mod_authn_dbi.c] DBI Connection Failed. dbi_conn_new returned NULL. Insufficient memory or invalid DBD?");
            rv = APR_EGENERAL;
            /*
             * modules/ssl/ssl_engine_log.c:103
             *  said this was okay. so i do it.
             */
            exit(1);
        }
        else {
            dbi_conn_set_option(myres->conn, "host", host);
            dbi_conn_set_option(myres->conn, "username", user);
            dbi_conn_set_option(myres->conn, "password", pwd);
            dbi_conn_set_option(myres->conn, "dbname", name);

            if (dbi_conn_connect(myres->conn) != 0) {
                err_num = dbi_conn_error(myres->conn, &err_str);
                /* Connetion Failed */
                ap_log_perror(APLOG_MARK, APLOG_ERR, 0, r,
                              "[mod_authn_dbi.c] DBI Connection to %s://%s@%s/%s Failed. Error: (%d) %s",
                              driver, user, host, name, err_num, err_str);
                rv = APR_EGENERAL;
            }
            else {
                ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, r,
                              "[mod_authn_dbi.c] Connection was created sucessfully");
            }
        }
        *resource = myres;
    }
    else {
        /* Error -- we have too many TOTAL DBI Connections. Maybe a Evil User trying to hurt our system? */
        ap_log_perror(APLOG_MARK, APLOG_ERR, 0, r,
                      "[mod_authn_dbi.c] DBI Connection Failed. Hard Max Limit of %d Connections has been reached",
                      DBI_HARD_MAX_CONNS);
        /* we didn't create a new connection! */
        dbi_conn_count--;
        rv = APR_EGENERAL;
    }
    return rv;
}

static apr_status_t safe_dbi_kill_conn(void *resource, void *params,
                                       apr_pool_t * pool)
{
    dbi_rest *res = resource;

    dbi_conn_close(res->conn);

    dbi_conn_count--;

    return APR_SUCCESS;
}

dbi_config *create_new_conf(conn_id conn_id, apr_pool_t * p)
{
    dbi_config *conf;
    conf = (dbi_config *) apr_pcalloc(p, sizeof(dbi_config));
    if (conf == NULL) {
        return NULL;
    }
    conf->rec.dbi_name = DFLT_DBI_NAME;
    conf->rec.dbi_driver = DFLT_DBI_DRIVER;
    conf->rec.dbi_host = DFLT_DBI_HOST;
    conf->rec.dbi_user = DFLT_DBI_USER;
    conf->rec.dbi_pass = DFLT_DBI_PASS;
    conf->rec.dbi_table = DFLT_DBI_TABLE;
    conf->rec.username_field = DFLT_USERNAME_FIELD;
    conf->rec.password_field = DFLT_PASSWORD_FIELD;
    conf->rec.password_query = DFLT_PASSWORD_QUERY;
    conf->rec.password_format = DFLT_PASSWORD_FORMAT;
    conf->rec.isactive_field = DFLT_ACTIVE_FIELD;
    conf->rec.conn_min = DFLT_CONN_MIN;
    conf->rec.conn_soft = DFLT_CONN_SOFT;
    conf->rec.conn_max = DFLT_CONN_MAX;
    conf->rec.conn_ttl = DFLT_CONN_TTL;
    conf->rec.options = DFLT_OPTIONS;
    apr_hash_set(authn_dbi_config, conn_id, APR_HASH_KEY_STRING, conf);
    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p,
                  "[mod_authn_dbi.c] Creating Config for %s", conn_id);
    return conf;
}

static apr_status_t get_or_create_dbi_conf(const char *conn_id,
                                           apr_pool_t * p,
                                           dbi_config ** confname)
{
    dbi_config *temp;
    unsigned int c;

    /* some sanity checks on conn_id..limits are liberal and are more or less random */
    if (strlen(conn_id) > 255) {
        return APR_EGENERAL;
    }
    for (c = 0; c < strlen(conn_id); c++) {
        if (conn_id[c] < ' ') {
            return APR_EGENERAL;
        }
    }
    temp = apr_hash_get(authn_dbi_config, conn_id, APR_HASH_KEY_STRING);
    if (temp == NULL) {
        /* no such server yet... */
        temp = create_new_conf(conn_id, p);
    }
    *confname = temp;
    return APR_SUCCESS;
}

conn_id encap_conn_id(cmd_parms * cmd, const char *conn_id)
{
    /* XXX: this will be used to allow configuration in htaccess */
    return conn_id;
}

#define QUERYSTRING_MAGIC_CHAR '&'
#define QUERYSTRING_LEFT_DELIM_CHAR '{'
#define QUERYSTRING_RIGHT_DELIM_CHAR '}'

#define EMPTY_VAR ""            /* do NOT set this to NULL! */

/* with a little help from ap_resolve_env() ;) */
const char *populate_querystring(request_rec * r, const char *querystring,
                                 dbi_config * conf, dbi_dconfig * dconf,
                                 dbi_rest * dbi_res, const char *user,
                                 const char *pw, const char *realm)
{

    char tmp[MAX_STRING_LEN];   /* 8 KByte should be enough for everyone :) */
    const char *s, *e;
    char *p;
    tmp[0] = '\0';
    int written = 0;

    if (!(s = ap_strchr_c(querystring, QUERYSTRING_MAGIC_CHAR)))
        return querystring;

    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, r->pool,
                  "[mod_authn_dbi.c] Querystring to populate: %s",
                  querystring);
    do {
        written += (s - querystring);
        //  MAX_STRING_LEN-1 so we still have room for a null at the end
        if (written >= MAX_STRING_LEN-1) {
            ap_log_perror(APLOG_MARK, APLOG_ERR, 0, r->pool,
                          "[mod_authn_dbi.c] Populated string would exceed %d bytes",
                          MAX_STRING_LEN);
            return NULL;
        }
        strncat(tmp, querystring, s - querystring);

        if ((s[1] == QUERYSTRING_LEFT_DELIM_CHAR)
            && (e = ap_strchr_c(s, QUERYSTRING_RIGHT_DELIM_CHAR))) {
            const char *e2 = e;
            char *var;
            p = NULL;
            querystring = e + 1;
            e = NULL;
            var = apr_pstrndup(r->pool, s + 2, e2 - (s + 2));

            if (!strcasecmp(var, "GivenPassword")) {
                e = (pw ? pw : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "GivenUsername")) {
                e = (user ? user : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "Realm")) {
                e = (realm ? realm : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "RequestURI")) {
                e = (r->uri ? r->uri : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "PathInfo")) {
                e = (r->path_info ? r->path_info : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "RequestArgs")) {
                e = (r->args ? r->args : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "RequestHostname")) {
                e = (r->hostname ? r->hostname : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "RequestFile")) {
                e = (r->filename ? r->filename : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "Name")) {
                e = (conf->rec.dbi_name ? conf->rec.dbi_name : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "AuthName")) {
                e = (ap_auth_name(r) ? ap_auth_name(r) : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "RemoteIP")) {
                e = (r->connection->remote_ip ? r->connection->remote_ip : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "RemoteHost")) {
                e = ap_get_remote_host(r->connection, r->per_dir_config, REMOTE_NOLOOKUP, NULL);
                (e?e : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "LocalIP")) {
                e = (r->connection->local_ip ? r->connection->local_ip : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "ConfigHostname")) {
                e = (r->server->server_hostname ? r->server->
                     server_hostname : EMPTY_VAR);
            }

            /* Everything but the variable values representing fieldnames and tables gets
             * escaped according to the selected driver */

            if (e != NULL) {
                p = strdup(e);
                dbi_driver_quote_string(dbi_conn_get_driver(dbi_res->conn),
                                        &p);
            }

            if (!strcasecmp(var, "UsernameField")) {
                e = (conf->rec.username_field ? conf->rec.
                     username_field : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "PasswordField")) {
                e = (conf->rec.password_field ? conf->rec.
                     password_field : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "IsActiveField")) {
                e = (conf->rec.isactive_field ? conf->rec.
                     isactive_field : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "Table")) {
                e = (conf->rec.dbi_table ? conf->rec.dbi_table : EMPTY_VAR);
            }

            if (e == NULL) {
                ap_log_perror(APLOG_MARK, APLOG_WARNING, 0, r->pool,
                              "[mod_authn_dbi.c] Unknown variable: %s", var);
                return NULL;
            }

            if (p == NULL) {
                p = strdup(e);
            }
            written += strlen(p);
            if (written >= MAX_STRING_LEN-1) {
                ap_log_perror(APLOG_MARK, APLOG_ERR, 0, r->pool,
                              "[mod_authn_dbi.c] Populated string would exceed %d bytes",
                              MAX_STRING_LEN);
                free(p);
                return NULL;
            }
            strcat(tmp, p);
            free(p);

        }
        else {
            ap_log_perror(APLOG_MARK, APLOG_WARNING, 0, r->pool,
                          "[mod_authn_dbi.c] Invalid querystring");
            return NULL;

        };

    } while ((s = ap_strchr_c(querystring, QUERYSTRING_MAGIC_CHAR)));

    strcat(tmp, querystring);
    written += strlen(querystring);
    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, r->pool,
                  "[mod_authn_dbi.c] Populated result: \"%s\" / %d chars written",
                  apr_pstrdup(r->pool, tmp), written);

    return apr_pstrdup(r->pool, tmp);
}

static const char *set_dbi_switch_conf(cmd_parms * cmd, void *config,
                                       const char *conn_id, const char *value)
{
    apr_ssize_t pos = (apr_ssize_t) cmd->info;
    dbi_config *temp;
    if ((get_or_create_dbi_conf
         (encap_conn_id(cmd, conn_id), cmd->pool, &temp)) == APR_SUCCESS) {

        /* Overwriting an existing value technically is a memory leak, since the pconf pool is only
         * destroyed at the termination of the whole apache process. Otoh, when processing htaccess,
         * we get handed the request-pool instead which is freed afterwards, so we should be fine. */

        switch (pos) {
        case CONF_DBI_DRIVER:
            temp->rec.dbi_driver = value;
            break;
        case CONF_DBI_HOST:
            temp->rec.dbi_host = value;
            break;
        case CONF_DBI_USERNAME:
            temp->rec.dbi_user = value;
            break;
        case CONF_DBI_PASSWORD:
            temp->rec.dbi_pass = value;
            break;
        case CONF_DBI_NAME:
            temp->rec.dbi_name = value;
            break;
        case CONF_DBI_TABLE:
            temp->rec.dbi_table = value;
            break;
        case CONF_DBI_USERNAME_FIELD:
            temp->rec.username_field = value;
            break;
        case CONF_DBI_PASSWORD_FIELD:
            temp->rec.password_field = value;
            break;
        case CONF_DBI_PASSWORD_FORMAT:
            if (!strcasecmp(value, "Plain")) {
                temp->rec.password_format = plain;
            }
            else if (!strcasecmp(value, "Apr")) {
                temp->rec.password_format = apr;
            }
            else if (!strcasecmp(value, "AprDigest")) {
                temp->rec.password_format = aprdig;
            }
            else {
                return apr_pstrcat(cmd->temp_pool,
                                   "Invalid password format: ", value, NULL);
            }
            break;
        case CONF_DBI_PASSWORD_QUERY:
            temp->rec.password_query = value;
            break;
        case CONF_DBI_IS_ACTIVE_FIELD:
            temp->rec.isactive_field = value;
            break;
        case CONF_DBI_CONN_MIN:
            temp->rec.conn_min = atoi(value);
            break;
        case CONF_DBI_CONN_SOFTMAX:
            temp->rec.conn_soft = atoi(value);
            break;
        case CONF_DBI_CONN_HARDMAX:
            temp->rec.conn_max = atoi(value);
            break;
        case CONF_DBI_CONN_TTL:
            temp->rec.conn_ttl = atoi(value);
            break;
        default:
            // unknown config directive?
            break;
        }
    }
    return NULL;
}

static const char *set_dbi_options(cmd_parms * cmd, void *config,
                                   const char *conn_id, const char *option)
{
    dbi_config *temp;
    if ((get_or_create_dbi_conf
         (encap_conn_id(cmd, conn_id), cmd->pool, &temp)) == APR_SUCCESS) {

        if (!strcasecmp(option, "AllowEmptyPasswords")) {
            temp->rec.options |= DBI_OPTION_EMPTY_PASSWORDS;
        }
        else {
            return apr_pstrcat(cmd->temp_pool,
                               "Unknown option passed do AuthDbiOptions for config ",
                               conn_id,
                               " passed to AuthnDbiOptions : ", option, NULL);
        }
    }
    return NULL;
}

static const char *set_dbi_driverdir(cmd_parms * cmd, void *config,
                                     const char *field)
{
    if (dbi_global_config.driverdir != NULL) {
        ap_log_perror(APLOG_MARK, APLOG_WARNING, 0, cmd->pool,
                      "[mod_authn_dbi.c] Overwriting previous AuthnDbiDriver value with new value %s",
                      dbi_global_config.driverdir);
    }
    dbi_global_config.driverdir = field;
    return NULL;
}

static const char *set_ddbi_conn(cmd_parms * cmd, void *config,
                                 const char *field)
{
    dbi_config *conf;

    /* we dont use get_or_create_dbi_conf here because we just look, and don't touch */
    if ((conf = apr_hash_get(authn_dbi_config, field, APR_HASH_KEY_STRING))) {
        ((dbi_dconfig *) config)->id = field;
    }
    else {
        ap_log_perror(APLOG_MARK, APLOG_ERR, 0, cmd->pool,
                      "[mod_authn_dbi.c] Unknown configuration %s", field);
    }
    return NULL;
}

static const command_rec authn_dbi_cmds[] = {
    /* global config items */
    AP_INIT_TAKE2("AuthnDbiDriver", set_dbi_switch_conf,
                  (void *) CONF_DBI_DRIVER, RSRC_CONF,
                  "The DBI Driver"),
    AP_INIT_TAKE1("AuthnDbiDriverDir", set_dbi_driverdir,
                  (void *) CONF_DBI_DRIVER_DIR, RSRC_CONF,
                  "The directory containing the DBI drivers"),
    AP_INIT_TAKE2("AuthnDbiHost", set_dbi_switch_conf, (void *) CONF_DBI_HOST,
                  RSRC_CONF,
                  "The host for the database connection"),
    AP_INIT_TAKE2("AuthnDbiUsername", set_dbi_switch_conf,
                  (void *) CONF_DBI_USERNAME, RSRC_CONF,
                  "The username for the database connection"),
    AP_INIT_TAKE2("AuthnDbiPassword", set_dbi_switch_conf,
                  (void *) CONF_DBI_PASSWORD, RSRC_CONF,
                  "The password for the database connection"),
    AP_INIT_TAKE2("AuthnDbiName", set_dbi_switch_conf, (void *) CONF_DBI_NAME,
                  RSRC_CONF,
                  "The name of the database containing the tables"),
    AP_INIT_TAKE2("AuthnDbiTable", set_dbi_switch_conf,
                  (void *) CONF_DBI_TABLE, RSRC_CONF,
                  "The name of the table containing the usernames and password hashes"),
    AP_INIT_TAKE2("AuthnDbiUsernameField", set_dbi_switch_conf,
                  (void *) CONF_DBI_USERNAME_FIELD, RSRC_CONF,
                  "The table field that contains the username"),
    AP_INIT_TAKE2("AuthnDbiPasswordField", set_dbi_switch_conf,
                  (void *) CONF_DBI_PASSWORD_FIELD,
                  RSRC_CONF,
                  "The table field that contains the password"),
    AP_INIT_TAKE2("AuthnDbiPasswordQuery", set_dbi_switch_conf,
                  (void *) CONF_DBI_PASSWORD_QUERY,
                  RSRC_CONF,
                  "The SQL query to pick the password field from"),
    AP_INIT_TAKE2("AuthnDbiPasswordFormat", set_dbi_switch_conf,
                  (void *) CONF_DBI_PASSWORD_FORMAT,
                  RSRC_CONF,
                  "The format the password is saved as (Plain, Apr, or AprDigest)"),
    AP_INIT_TAKE2("AuthnDbiIsActiveField", set_dbi_switch_conf,
                  (void *) CONF_DBI_IS_ACTIVE_FIELD,
                  RSRC_CONF,
                  "The table field that contains the username"),
    AP_INIT_TAKE2("AuthnDbiConnMin", set_dbi_switch_conf,
                  (void *) CONF_DBI_CONN_MIN, RSRC_CONF,
                  "The Minimum Number of Database Connections"),
    AP_INIT_TAKE2("AuthnDbiConnSoftMax", set_dbi_switch_conf,
                  (void *) CONF_DBI_CONN_SOFTMAX, RSRC_CONF,
                  "The Soft Maximum Number of Database Connections"),
    AP_INIT_TAKE2("AuthnDbiConnHardMax", set_dbi_switch_conf,
                  (void *) CONF_DBI_CONN_HARDMAX, RSRC_CONF,
                  "The Hard Maximum Number of Database Connections"),
    AP_INIT_TAKE2("AuthnDbiConnTTL", set_dbi_switch_conf,
                  (void *) CONF_DBI_CONN_TTL, RSRC_CONF,
                  "The Database Pool Time To Live for Each Connection."),
    AP_INIT_ITERATE2("AuthnDbiOptions", set_dbi_options,
                     (void *) CONF_DBI_OPTIONS, RSRC_CONF,
                     "A list of options for this configuration"),

    /* per auth section */
    AP_INIT_TAKE1("AuthnDbiServerConfig", set_ddbi_conn, NULL, OR_AUTHCFG,
                  "The name of the configuration to use for this section"),
    {NULL}
};

module AP_MODULE_DECLARE_DATA authn_dbi_module;

static void *create_authn_dbi_dir_config(apr_pool_t * p, char *d)
{
    dbi_dconfig *conf;
    if (d == NULL) {
        return NULL;
    }
    conf = (dbi_dconfig *) apr_pcalloc(p, sizeof(dbi_dconfig));
    if (conf) {
        conf->id = NULL;
    }
    return conf;
}

static apr_status_t safe_dbi_rel_server(apr_reslist_t * dbi_pool,
                                        dbi_rest * server, request_rec * r)
{
    apr_status_t rv;
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "[mod_authn_dbi.c] Returning Server Connection to DBI Pool");
    if(dbi_pool) {
	rv = apr_reslist_release(dbi_pool, (void **) server);
	return rv;
    } 
    else {
	/* XXX: This can not really happen, but just to be on the safe side... */
	return APR_EGENERAL;
    }
}

static int safe_dbi_query(dbi_rest * mydbi_res, dbi_result * res,
                          request_rec * r, const char *query)
{
    int err_num = 0;
    const char *err_str;
    int error = 1;

    *res = (dbi_result) dbi_conn_query(mydbi_res->conn, query);

    /* logging complete sql queries is bad. I personaly 
     * uncomment this for some debuging... but even 
     * APLOG_DEBUG isn't good for this.
     */
    /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
     *                   "[mod_authn_dbi.c] SQL Query: %s", query);
     */


    if (res == NULL) {
        err_num = dbi_conn_error(mydbi_res->conn, &err_str);
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                      "[mod_authn_dbi.c] SQL Query Failed.  DBI said: (%d) %s",
                      err_num, err_str);
    }
    else {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                      "[mod_authn_dbi.c] Query Result is good.");
        error = 0;
    }
    return error;
}

static authn_status check_dbi_pw(request_rec * r, const char *user,
                                 const char *password)
{
    authn_status ARV = AUTH_DENIED;
    dbi_config *conf;
    dbi_dconfig *dconf = ap_get_module_config(r->per_dir_config,
                                              &authn_dbi_module);
    char *dbi_pass = NULL;
    const char *query;
    dbi_result result;
    dbi_rest *dbi_res;
    apr_status_t aprval;
    char crypt_pass[34];

    conf = apr_hash_get(authn_dbi_config, dconf->id, APR_HASH_KEY_STRING);
    if (conf == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "[mod_authn_dbi.c] - Server Config for \"%s\" was not found",
                      dconf->id);
        return AUTH_GENERAL_ERROR;
    }

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "[mod_authn_dbi.c] Attempting to Acquire DBI Connection");
    if(!conf->pool || apr_reslist_acquire(conf->pool, (void **) &dbi_res) != APR_SUCCESS) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
		      "[mod_authn_dbi.c] Failed to Acquire DBI Connection");
	return AUTH_GENERAL_ERROR;
    }
    /* make the query to get the user's password */
    if (conf->rec.isactive_field) {
        if (conf->rec.password_query == NULL) {
            query =
                "SELECT &{PasswordField} FROM &{Table} WHERE &{UsernameField}=&{GivenUsername} AND &{IsActiveField}!=0 LIMIT 0,1";
        }
        else {
            query = conf->rec.password_query;
        }
    }
    else {
        if (conf->rec.password_query == NULL) {
            query =
                "SELECT &{PasswordField} FROM &{Table} WHERE &{UsernameField}=&{GivenUsername} LIMIT 0,1";
        }
        else {
            query = conf->rec.password_query;
        }
    }

    /* perform the query */
    if ((query =
         populate_querystring(r, query, conf, dconf, dbi_res, user, password,
                              NULL))
        && safe_dbi_query(dbi_res, &result, r, query) == 0) {
        /* store the query result */
        if (dbi_result_next_row(result)
            && dbi_result_get_numrows(result) == 1) {

            dbi_pass =
                dbi_result_get_string_copy(result, conf->rec.password_field);
            if ((dbi_pass == NULL) || (strcmp(dbi_pass, "ERROR") == 0)) {
                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                              "[mod_authn_dbi.c] - libdbi returned an error when retrieving the password field from the query result");
                ARV = AUTH_GENERAL_ERROR;
            }
            else if (!strcmp(dbi_pass, "!") || !strcmp(dbi_pass, "*")) {
                /* Basically, we just suppress a warning with this and it could just as well be achieved with
                 * the IsActive field. we'll keep this in the holy name of blind GNU featureism :-} */
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                              "Account for user %s is disabled.", user);
                ARV = AUTH_DENIED;
            }
            else if ((conf->rec.options & DBI_OPTION_EMPTY_PASSWORDS)
                     && !strcmp(dbi_pass, DBI_EMPTY_PASSWORD)) {
                /* This user is configured to be allowed to login without a password (or any other password) */
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                              "Accepting user %s without password check.",
                              user);
                ARV = AUTH_GRANTED;
            }
            else {
                /* XXX: what about portability!? are any special precautions needed for win32/netware/ebdic ? */
                switch (conf->rec.password_format) {

                case plain:
                    if (strcmp(dbi_pass, password) != 0) {
                        ARV = AUTH_DENIED;
                    }
                    else {
                        ARV = AUTH_GRANTED;
                    }
                    break;

                case apr:
                    /* 
                     * This handles the password through the aprutil library.
                     * At the moment, the supported formats are crypt (on *nix only), SHA1 and MD5.
                     * Other formats will transparently be supported as they become
                     * available in libaprutil.
                     *
                     * Examples are:
                     * crypt        O8D24p2LCO7PA
                     * md5          $apr1$SvzPV/..$CJl3oQ/ko4Tq5eg6L2Fk..
                     * sha1         {SHA}C+7Hteo/D9vJXQ3UfzxbwnXaijM=
                     */

#ifndef SKIP_APR_PWCHECK
                    /* Define SKIP_APR_PWCHECK to disable this check if apr-util supports new algorithms we don't yetSKIP_APR_PWCHECK
                     * know about. Should be used as a workaround only since it might open the door to DoS attacks if specially
                     * crafted entries are passed to apr_password_validate.
                     */
                    {
                        int c;

                        c = strlen(dbi_pass);
                        /* we do some checks to keep libc from making wrong assumptions and ourselves from segfaulting */
                        if (!((c == 13) && (apr_isalnum(dbi_pass[0]) || (dbi_pass[0] == '.') || (dbi_pass[0] == '.')) 
                            && (apr_isalnum(dbi_pass[1]) || (dbi_pass[1] == '.') || (dbi_pass[1] == '.'))) && /* crypt */
                            !((c == 37)) &&       /* not much to be done for md5, apr-util is old enough to look after itself */
                            (0 != strncmp(dbi_pass, APR_SHA1PW_ID, strlen(APR_SHA1PW_ID)))) { /* sha1 */     
                            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                                          "[mod_authn_dbi.c] Invalid Apr-format entry or unsupported format in password field of user %s in config %s",
                                          user, conf->rec.dbi_name);
                            ARV = AUTH_GENERAL_ERROR;
                            break;
                        }
                    }
#endif
                    if ((aprval =
                         apr_password_validate(password,
                                               dbi_pass)) == APR_SUCCESS) {
                        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                                      "[mod_authn_dbi.c] Successully authenticated user %s",
                                      user);
                        ARV = AUTH_GRANTED;
                    }
                    else {
                        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                                      "[mod_authn_dbi.c] Authentication for user %s failed",
                                      user);
                        ARV = AUTH_DENIED;
                    }
                    break;

                case aprdig:
                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                                  "[mod_authn_dbi.c] - AprDig format is not supported for Basic Authentication.");
                    ARV = AUTH_GENERAL_ERROR;
                    break;

                default:
                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                                  "[mod_authn_dbi.c] - Internal error. Unimplemented AuthnDbiPasswordFormat was configured.");
                    ARV = AUTH_GENERAL_ERROR;
                }
            }
        }
        else {
            if (dbi_result_get_numrows(result) == 0) {
                ARV = AUTH_USER_NOT_FOUND;
            }
            else {
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                              "[mod_authn_dbi.c] %lu row(s) was not returned by dbi_result_get_numrows(result)",
                              (unsigned long) dbi_result_get_numrows(result));
                ARV = AUTH_DENIED;
            }
        }
        dbi_result_free(result);
    }
    else {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                      "[mod_authn_dbi.c] Query Failed!");
        ARV = AUTH_GENERAL_ERROR;
    }
    if (dbi_pass != NULL) {
        free(dbi_pass);
    }
    safe_dbi_rel_server(conf->pool, dbi_res, r);
    return ARV;
}

static authn_status get_dbi_realm_hash(request_rec * r, const char *user,
                                       const char *realm, char **rethash)
{
    dbi_config *conf;
    authn_status ARV = AUTH_DENIED;
    dbi_dconfig *dconf = ap_get_module_config(r->per_dir_config,
                                              &authn_dbi_module);
    char *dbi_hash;
    char *dbi_pass = NULL;
    const char *query;
    dbi_result result;
    dbi_rest *dbi_res;

    conf = apr_hash_get(authn_dbi_config, dconf->id, APR_HASH_KEY_STRING);
    if (conf == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "[mod_authn_dbi.c] - Server Config for \"%s\" was not found",
                      dconf->id);
        return AUTH_GENERAL_ERROR;
    }

    if ((conf->rec.password_format != plain)
        && (conf->rec.password_format != aprdig)) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "[mod_authn_dbi.c] - Digest Authentication with AuthnDbiPasswordFormat other than 'Plain' or 'AprDigest' is only supported on quantum computers");
        return AUTH_GENERAL_ERROR;
    }

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "[mod_authn_dbi.c] Attempting to Acquire DBI Connection");
    if(!conf->pool || apr_reslist_acquire(conf->pool, (void **) &dbi_res) != APR_SUCCESS) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
		      "[mod_authn_dbi.c] Failed to Acquire DBI Connection");
	return AUTH_GENERAL_ERROR;
    }
    
    /* make the query to get the user's password */

    if (conf->rec.isactive_field) {
        if (conf->rec.password_query == NULL) {
            query =
                "SELECT &{PasswordField} FROM &{Table} WHERE &{UsernameField}=&{GivenUsername} AND &{IsActiveField}!=0 LIMIT 0,1";
        }
        else {
            query = conf->rec.password_query;
        }
    }
    else {
        if (conf->rec.password_query == NULL) {
            query =
                "SELECT &{PasswordField} FROM &{Table} WHERE &{UsernameField}=&{GivenUsername} LIMIT 0,1";
        }
        else {
            query = conf->rec.password_query;
        }
    }

    /* perform the query */
    if ((query =
         populate_querystring(r, query, conf, dconf, dbi_res, user, NULL,
                              realm))
        && safe_dbi_query(dbi_res, &result, r, query) == 0) {
        /* store the query result */
        if (dbi_result_next_row(result)
            && dbi_result_get_numrows(result) == 1) {
            dbi_pass =
                dbi_result_get_string_copy(result, conf->rec.password_field);

            if ((dbi_pass == NULL) || (strcmp(dbi_pass, "ERROR") == 0)) {
                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                              "[mod_authn_dbi.c] - libdbi returned an error when retrieving the password field from the query result");
                ARV = AUTH_GENERAL_ERROR;
            }
            else if (!strcmp(dbi_pass, "!") || !strcmp(dbi_pass, "*")) {
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                              "Account for user %s is disabled.", user);
                ARV = AUTH_DENIED;
            }
            else if ((conf->rec.options & DBI_OPTION_EMPTY_PASSWORDS)
                     && !strcmp(dbi_pass, DBI_EMPTY_PASSWORD)) {
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                              "Accepting user %s without password check.",
                              user);
                ARV = AUTH_GRANTED;
            }
            else {

                switch (conf->rec.password_format) {
                case plain:
                    dbi_hash = (char *) ap_md5(r->pool,
                                               (const unsigned char *)
                                               apr_pstrcat(r->pool, user, ":",
                                                           realm, ":",
                                                           dbi_pass, NULL));

                    *rethash = dbi_hash;
                    ARV = AUTH_USER_FOUND;
                    break;

                case aprdig:
                    if (strlen(dbi_pass) != 32) {
                        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                                      "[mod_authn_dbi.c]  Invalid Apr-format entry or unsupported format in password field of user %s in config %s",
                                      user, conf->rec.dbi_name);
                        ARV = AUTH_GENERAL_ERROR;
                    }
                    else {
                        *rethash = apr_pstrdup(r->pool, dbi_pass);
                        ARV = AUTH_USER_FOUND;
                    }
                    break;
                default:
                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                                  "[mod_authn_dbi.c] Unsupported password format in digest authentication. This is an internal error of mod_authn_dbi");
                    ARV = AUTH_GENERAL_ERROR;
                }
            }
        }
        else {
            if (dbi_result_get_numrows(result) == 0) {
                ARV = AUTH_USER_NOT_FOUND;
            }
            else {
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                              "[mod_authn_dbi.c] %lu row(s) was not returned by dbi_result_get_numrows(result)",
                              (unsigned long) dbi_result_get_numrows(result));
                ARV = AUTH_DENIED;
            }
        }
        dbi_result_free(result);
    }
    else {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                      "[mod_authn_dbi.c] Query Failed!");
        ARV = AUTH_GENERAL_ERROR;
    }
    if (dbi_pass != NULL) {
        free(dbi_pass);
    }
    safe_dbi_rel_server(conf->pool, dbi_res, r);
    return ARV;
}

static const authn_provider authn_dbi_provider = {
    &check_dbi_pw,
    &get_dbi_realm_hash
};

static apr_status_t init_authn_dbi_config(apr_pool_t * pconf,
                                          apr_pool_t * plog,
                                          apr_pool_t * ptemp)
{
    apr_status_t rv = APR_SUCCESS;
    /* create our globalish config var */
    authn_dbi_config = apr_hash_make(pconf);
    return rv;
}

static apr_status_t kill_dbi(void *p)
{
    apr_status_t rv = APR_SUCCESS;
    apr_hash_index_t *idx;
    char *key;
    dbi_config *val;
    apr_ssize_t len;

    for (idx = apr_hash_first((apr_pool_t *) p, authn_dbi_config); idx;
         idx = apr_hash_next(idx)) {
        apr_hash_this(idx, (void *) &key, &len, (void *) &val);
	if(val->pool)
	    apr_reslist_destroy(val->pool);
    }
    dbi_shutdown();

    return rv;
}


static apr_status_t init_authn_dbi(apr_pool_t * p, apr_pool_t * plog,
                                   apr_pool_t * ptemp, server_rec * s)
{
    apr_status_t rv = APR_SUCCESS;
    int rval;
    dbi_driver dbi_driver;
    void *data;
    apr_hash_index_t *idx;
    char *key;
    dbi_config *val;
    apr_ssize_t len;
    const char *userdata_key = "mod_authn_dbi_init";

    apr_pool_userdata_get(&data, userdata_key, s->process->pool);

    if (!data) {
        apr_pool_userdata_set((const void *) 1, userdata_key,
                              apr_pool_cleanup_null, s->process->pool);
        return OK;
    }
    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p,
                  "[mod_authn_dbi.c] Running DBI init Code");
    if ((rval = dbi_initialize(dbi_global_config.driverdir)) > 0) {
        if (dbi_global_config.driverdir == NULL) {
            ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p,
                          "[mod_authn_dbi.c] Initialization of libdbi found %d drivers in default driver directory",
                          rval);
        }
        else {
            ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p,
                          "[mod_authn_dbi.c] Initialization of libdbi found %d drivers in directory %s",
                          rval, dbi_global_config.driverdir);
        }
        if (s->loglevel >= APLOG_DEBUG) {
            dbi_driver = NULL;
            while ((dbi_driver = dbi_driver_list(dbi_driver)) != NULL) {
                ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p,
                              "[mod_authn_dbi.c] Driver '%s' was loaded.",
                              dbi_driver_get_name(dbi_driver));
            }
        }
    }
    else {                      /* An error was returned or libdbi found 0 drivers */
        if (dbi_global_config.driverdir == NULL) {
            ap_log_perror(APLOG_MARK, APLOG_EMERG, 0, p,
                          "[mod_authn_dbi.c] - Initlialization of libdbi with default driver directory failed");
        }
        else {
            ap_log_perror(APLOG_MARK, APLOG_EMERG, 0, p,
                          "[mod_authn_dbi.c] - Initlialization of libdbi with AuthnDbiDriverDir %s failed",
                          dbi_global_config.driverdir);
        }
        return APR_EDSOOPEN;
    }

    /* loop the hashed config stuff... */
    for (idx = apr_hash_first(p, authn_dbi_config); idx;
         idx = apr_hash_next(idx)) {
        apr_hash_this(idx, (void *) &key, &len, (void *) &val);
        if(apr_reslist_create(&val->pool, val->rec.conn_min,       /* hard minimum */
                           val->rec.conn_soft,  /* soft maximum */
                           val->rec.conn_max,   /* hard maximum */
                           val->rec.conn_ttl,   /* Time to live -- dbi server might override/disconnect! */
                           safe_dbi_new_conn,   /* Make a New Connection */
                           safe_dbi_kill_conn,  /* Kill Old Connection */
			      (void *) &val->rec, p) != APR_SUCCESS)
	    ap_log_perror(APLOG_MARK, APLOG_EMERG, 0, p,
                          "[mod_authn_dbi.c] - Creation of dbi connection pool failed for config set %s",
                          key);
	/* XXX: If no new connection could be opened, the reslist will not
	   contain any resources and reslist_create will fail, giving 
	   val->pool the value NULL and making it unusable. The configuration 
	   set is not really recoverable in this case. For now we log this condition
	   and the authentication functions protect themselves against this 
	   condition. A much better way would be to free the configuration
           set completely. But to do this other changes are necessary, so for
           now this is basically a workaround. This has to be fixed before
	   1.0 */
        apr_hash_set(authn_dbi_config, key, APR_HASH_KEY_STRING, val);
    }

    apr_pool_cleanup_register(p, p, kill_dbi, apr_pool_cleanup_null);

    ap_add_version_component(p, "mod_authn_dbi/" MOD_AUTHN_DBI_VERSION);

    return rv;
}

static void register_hooks(apr_pool_t * p)
{
    ap_hook_pre_config(init_authn_dbi_config, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_post_config(init_authn_dbi, NULL, NULL, APR_HOOK_MIDDLE);
    ap_register_provider(p, AUTHN_PROVIDER_GROUP, "dbi", "0",
                         &authn_dbi_provider);
}

module AP_MODULE_DECLARE_DATA authn_dbi_module = {
    STANDARD20_MODULE_STUFF,
    create_authn_dbi_dir_config,        /* dir config creater */
    NULL,                               /* dir merger --- default is to override */
    NULL,                               /* server config creator */
    NULL,                               /* merge server config */
    authn_dbi_cmds,                     /* command apr_table_t */
    register_hooks                      /* register hooks */
};
